# -*- coding: utf-8 -*-
"""
Some SQLAlchemy specific field types.
"""
# -*- coding: utf-8 -*-
from rest_framework import fields


# TODO: Should build a url to the resource
class UriField(fields.CharField):
    """
    Represents a uri to the resource
    """

    def __init__(self, view_name=None, *args, **kwargs):
        self.view_name = view_name
        super(UriField, self).__init__(*args, **kwargs)


class EnumField(fields.ChoiceField):
    """
    Used for SQLAlchemy's Enum column type used in either mapping python Enum's, or
    a list of valid fields for the column.
    """

    def __init__(self, **kwargs):
        self.enum_class = kwargs.pop('enum_class')
        kwargs['choices'] = [(e.name, e.name) for e in self.enum_class]
        kwargs.pop('max_length', None)
        super(EnumField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        try:
            return self.enum_class[data]
        except KeyError:
            self.fail('invalid_choice', input=data)

    def to_representation(self, value):
        if not value:
            return None

        return value.name


class CharMappingField(fields.DictField):
    """
    Used for Postgresql HSTORE columns for storing key-value pairs.
    """
    child = fields.CharField(allow_null=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.test import SimpleTestCase
from rest_witchcraft import utils
from sqlalchemy import Column, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class GetPkModel(Base):
    __tablename__ = 'getpk'

    pk1 = Column(String(), primary_key=True)


class GetPkCompositeModel(Base):
    __tablename__ = 'getpkcomposite'

    pk1 = Column(String(), primary_key=True)
    pk2 = Column(String(), primary_key=True)


class TestModelRoutes(SimpleTestCase):

    def test_get_pks_returns_none_when_no_pk_found(self):

        pks = utils.get_primary_keys(GetPkModel, {})

        self.assertIsNone(pks)

    def test_get_pks_returns_none_when_pk_none(self):

        pks = utils.get_primary_keys(GetPkModel, {'pk1': None})

        self.assertIsNone(pks)

    def test_get_pks_returns_none_when_one_pk_not_found(self):

        pks = utils.get_primary_keys(GetPkCompositeModel, {'pk1': 'a'})

        self.assertIsNone(pks)

    def test_get_pks_returns_none_when_one_pk_is_none(self):

        pks = utils.get_primary_keys(GetPkCompositeModel, {'pk1': 'a', 'pk2': None})

        self.assertIsNone(pks)



# -*- coding: utf-8 -*-
import contextlib
import inspect
from collections import OrderedDict
from itertools import chain

import six
from django.utils.text import capfirst
from sqlalchemy import inspect as sa_inspect

try:
    suppress = contextlib.suppress
except AttributeError:

    @contextlib.contextmanager
    def suppress(*exceptions):
        """
        Suppresses given exceptions, a backport from py3 contextlib.suppress
        """
        try:
            yield
        except exceptions:
            pass


if six.PY2:

    def get_args(func):
        """
        Returns the names of the positional arguments for composite model inspection
        """
        return inspect.getargspec(func).args[1:]  # pylint:disable=deprecated-pragma
else:

    def get_args(func):
        """
        Returns the names of the positional arguments for composite model inspection
        """
        return list(inspect.signature(func).parameters.keys())[1:]


class composite_info(object):
    """
    A helper class that makes sqlalchemy composite model inspection easier
    """
    __slots__ = ('prop', 'properties', '_field_names')

    def __init__(self, composite):
        self._field_names = set()
        self.prop = composite.prop
        attrs = get_args(self.prop.composite_class.__init__)

        self.properties = {}
        for attr, prop, col in zip(attrs, self.prop.props, self.prop.columns):
            self.properties[attr] = _column_info(prop, col)

    @property
    def field_names(self):
        if not self._field_names:
            self._field_names.update(self.properties.keys())

            self._field_names = [attr for attr in self._field_names if not attr.startswith('_')]

        return self._field_names


class _column_info(object):
    __slots__ = ('property', 'column')

    def __init__(self, prop, column):
        self.property = prop
        self.column = column

    @property
    def field_kwargs(self):
        kwargs = {'label': capfirst(' '.join(self.property.key.split('_'))), 'help_text': self.property.doc}

        with suppress(AttributeError):
            enum_class = self.column.type.enum_class
            if enum_class is not None:
                kwargs['enum_class'] = self.column.type.enum_class
            else:
                kwargs['choices'] = self.column.type.enums

        with suppress(AttributeError):
            kwargs['max_digits'] = self.column.type.precision

        with suppress(AttributeError):
            kwargs['decimal_places'] = self.column.type.scale

        with suppress(AttributeError):
            kwargs['max_length'] = self.column.type.length

        kwargs['required'] = not self.column.nullable
        kwargs['allow_null'] = self.column.nullable

        return kwargs


class model_info_meta(type):
    _registry = {}

    def __call__(cls, model, *args, **kwargs):
        if model not in cls._registry:
            instance = super(model_info_meta, cls).__call__(model, *args, **kwargs)
            cls._registry[model] = instance

        return cls._registry[model]


class model_info(six.with_metaclass(model_info_meta)):
    """
    A helper class that makes sqlalchemy model inspection easier
    """

    __slots__ = ('mapper', 'properties', '_field_names', 'model_class', 'composites', 'primary_keys', 'relationships')

    def __init__(self, model):
        self.model_class = model
        self.mapper = sa_inspect(model)
        self._field_names = None
        self.properties = {}
        self.composites = {}
        self.relationships = {}
        self.primary_keys = OrderedDict()

        for col in self.mapper.primary_key:
            attr = self.mapper.get_property_by_column(col)
            self.primary_keys[attr.key] = _column_info(attr, col)

        for col in self.mapper.columns:
            attr = self.mapper.get_property_by_column(col)
            if attr.key not in self.primary_keys:
                self.properties[attr.key] = _column_info(attr, col)

        for composite in self.mapper.composites:
            self.composites[composite.key] = composite_info(getattr(model, composite.key))

        for relationship in self.mapper.relationships:
            self.relationships[relationship.key] = relationship

    @property
    def field_names(self):
        if not self._field_names:
            self._field_names = [
                attr
                for attr in chain(
                    self.primary_keys.keys(), self.properties.keys(), self.composites.keys(), self.relationships.keys()
                ) if not attr.startswith('_')
            ]

        return self._field_names


def get_primary_keys(model, kwargs):
    """
    Returns the primary key from a dictionary to be used in a sqlalchemy query.get() call
    """
    info = model_info(model)
    pks = []

    for attr, _ in info.primary_keys.items():
        pk = kwargs.get(attr)
        pks.append(pk)

    if len(pks) < 2:
        return next(iter(pks), None)

    if any(pk is None for pk in pks):
        return None

    return tuple(pks)



# -*- coding: utf-8 -*-
from rest_framework import mixins


class DestroyModelMixin(mixins.DestroyModelMixin):
    """
    Deletes a model instance
    """

    def perform_destroy(self, instance):
        session = self.get_session()
        session.delete(instance)
